/*
 *  Licensed Materials - Property of IBM
 *
 *  (C) COPYRIGHT IBM Corp. 1994, 1995, 1996  All rights reserved
 *
 *  US Government Users Restricted Rights - Use, duplication or
 *  disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
 *
 * Filename : 40
 * Author   : M. Flickner
 * Date     : 10/17/97 10:11:50
 * Version  : 1.8.3.3
 * Purpose  :
 *
 */

#include "Qbic.h"
#include "QbAbsCls.hpp"
#include "QbCreate.hpp"
#include "QbImgDat.hpp"
#include "QbQueryS.hpp"
#include "QbQueryC.hpp"
#include "QbEnv.h"


//
// define the functions to create instances of the classes
//
extern QbFeatureClass *NewQbColorFeatureClass() ;
extern QbFeatureClass *NewQbTextFeatureClass() ;
extern QbFeatureClass *NewQbTextureFeatureClass() ;
extern QbFeatureClass *NewQbColorHistogramFeatureClass() ;
extern QbFeatureClass *NewQbDrawFeatureClass() ;
extern QbFeatureClass *NewQbThumbnailClass() ;

static const struct {
  const char *featureName ; 
  QbFeatureClass* (*newFunction)() ;
}  FeatureClassTable[] = {
  "QbColorHistogramFeatureClass", NewQbColorHistogramFeatureClass,
  "QbDrawFeatureClass", NewQbDrawFeatureClass,
  "QbColorFeatureClass",  NewQbColorFeatureClass,
  "QbTextFeatureClass", NewQbTextFeatureClass,
  "QbTextureFeatureClass",  NewQbTextureFeatureClass, 
  NULL, NULL
} ;

//
// Function to list all features that are available for Qbic indexing
//
DMBEXPCP (const char **) QbListAllAvailableFeatures()
{
  int i ;
  static const char *featureNames[100] ; 
  for( i=0;  FeatureClassTable[i].featureName; i++ )  {
    featureNames[i] = FeatureClassTable[i].featureName;
  }
  return featureNames ; 
}

//
// Function to create a feature class given is name as a string
//
DMBEXPCP( QbFeatureClass *) QbCreateFeatureClass (const char *className )
{
  int i ;
  for( i=0;  FeatureClassTable[i].featureName; i++ )  {
    if( ! strcmp( className, FeatureClassTable[i].featureName ) )
      return FeatureClassTable[i].newFunction() ; 
  }
  return NULL ;
}

extern QbConnectClass *NewQbDbmConnectClass() ;

DMBEXPCP( QbConnectClass *) QbCreateConnectClass(const char *className )
{
  if( ! strcmp( className, "QbDbmConnectClass") )
    return NewQbDbmConnectClass();
  return NULL ;
}


DMBEXPCP( QbGenericDataClass *) QbCreateGenericDataClass(const char *className)
{
  if( ! strcmp( className, "QbGenericImageDataClass" ) )
    return new QbGenericImageDataClass ;
  return NULL ;
}

DMBEXPCP( QbParameterClass *) QbCreateParameterClass(const char *className)
{
    return new QbParameterClass ;
}

//
// create instances of query class
//
DMBEXPCP ( QbQueryClass *) QbCreateQueryClass( const char *className )
{
// we do not need client for extenders
  if( ! strcmp( className, "QbQueryServerClass" ) )
    return new QbQueryServerClass ;
  return NULL ;
}


static QbConnectClass *connection = NULL;

//
//To get things started you need to connect to a database.  To do this
//you need to know the class name for the QbConnectClass, the dataset
//name, and the mode, "r" or "w", you wish to connect with.  This is
//done once per address space in this implementation.  More
//sophisticated implementations are possible.
//
//I don't want caller deleting, disconnecting or connecting with the
//return value and I can enforce that using const.
//
DMBEXPCP(const QbConnectClass *) QbGetConnection(const char *modeArg,
                                      const char *dsNameArg,
                                      const char *connectClassNameArg)

{
  static const char *defaultConnectClassName = {"QbDbmConnectClass"} ;
  static const char *defaultDsName = {"QbicData"} ;
  int rc ;

  const char *dsName, *mode, *connectClassName ;

  if( connection ) return connection ; // only one connection per address space

  mode = Qbgetenv("QbicConnectMode") ;
  if( modeArg ) mode = modeArg ; // used passed mode

  dsName  = Qbgetenv("QbicDatabaseName") ; // get environment variable
  if( dsNameArg ) dsName = dsNameArg ; // used passed arg
  if( NULL == dsName ) dsName = defaultDsName ; // lastly default

  connectClassName = Qbgetenv("QbicConnectClass") ; // get env variable
  if( connectClassNameArg ) connectClassName = connectClassNameArg ;
  if( NULL == connectClassName )  connectClassName = defaultConnectClassName ;

  connection = QbCreateConnectClass( connectClassName ) ;
  rc = connection->Connect( dsName, mode ) ;
  if( rc ) {
    delete connection ;
    connection = NULL ;
  }
  return connection ;
}

//
// Process to Delete the connection
//
DMBEXPCP(int ) QbDestroyConnection( const QbConnectClass *connectArg )
{
  if( connectArg == connection ) {
    delete connection ;
    connection = NULL ;
  } else return -1 ;
  return 0 ;
}
