/*
 *  IBM Confidential
 *  
 *  Licensed Materials - Property of IBM
 *
 *  (C) COPYRIGHT IBM Corp. 1994-1998 All rights reserved
 *
 *  US Government Users Restricted Rights - Use, duplication or
 *  disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
 *
 * Filename : QbExampl.cpp
 * Date     : 12/2/98 09:59:13
 * Version  : 1.8
 * Author   : J. Hafner
 * Purpose  : Example class -- start from here to generate a new class from
 *            scratch
 *
 *
 */

static char sccsid[] = "@(#)QbExampl.cpp	1.8 12/2/98" ; 

static const char Copyright[] = {
  "Licensed Materials - Property of IBM\n"
  "(C) Copyright IBM Corp. 1994-1998  All rights Reserved\n"
  "US Government Users Restricted Rights - Use, duplication or\n"
  "disclosure restricted by GSA ADP Schedule Contract with IBM Corp." } ;

/* End of Common Block                                                  */

#include "QbExampl.hpp"
#include "QbImgDat.hpp" // this uses image features
#include "QbCatCls.hpp"

#ifdef INDEX_CODE
#include "QbIndex.hpp"
#endif

NEW_FUNCTION( QbFeatureClass, QbExampleFeatureClass )
NEW_FUNCTION( QbFeatureClass, QbExampleSubPartFeatureClass )

int QbExampleFeatureClass::isInitialized = 0 ;
QbParameterClass  QbExampleFeatureClass::defaultParameters ;

//
// Constructor sets up a conversion table one for all instances of the
// class.
//
QbExampleFeatureClass::QbExampleFeatureClass()
{
  if( ! isInitialized  ) {
    // do any one-time initialization here (including default parameters)
    // these are key-value pairs -- we fill in one generic one for kicks
    defaultParameters.SetFeatureCodes("Noop", ".") ;
    isInitialized  = True ;
  }
  // do any other initialization here (e.g., initializing feature vector)
  // if anything failed along the way, uncomment the following line
  // IamOK = False;
}
// Computes the size of the feature vector
unsigned long QbExampleFeatureClass::FeatureSize() {
  unsigned long size;
  size = 3*sizeof(long);
  return size;
}

//
// Converts the contents of a QbExampleFeatureClass to a bytestream
// Walk over the feature data you need to store in the database and
// put it into a datum (this example does our featureVector)
//
int QbExampleFeatureClass::ToByteString(QbDatumClass &data)
{
  unsigned char *buffer ;
  long *longPtr ;
  unsigned long size = FeatureSize();
  int i ;
  
  buffer = (unsigned char *)QbNew(size, "QbExampl::ToByteString");// get space
  if(!buffer) return -1;
  longPtr = (long *) buffer ; // start at the beginning
  for(i=0; i<3 ; i++) {
    *(longPtr++) = featureVector[i] ; // we should make this OS-independent
  }
  // convert the buffer to datum, and tell the datum that we
  // created the storage here
  data.Set( size, buffer, QbDatumClass::newArrayStorage ) ; 
  return 0 ;
}

//
// Set the state of a QbExampleFeature Class from a bytestream
// This is done when read from database (it inverts ToByteString)
//
int QbExampleFeatureClass::FromByteString(const QbDatumClass &data)
{
  unsigned char *buffer ;
  unsigned long size ;
  int i  ;
  long *longPtr ;

  data.Get( size, buffer ) ; // convert the data to string and get it's size
  if ( size != FeatureSize() ) { // didn't get right amount of bytes
    QbError(QERR_SIZE_MISMATCH, QERR_FATAL, "QbExampl", FeatureSize(), size);
    return -1 ;
  }      

  longPtr = (long *) buffer ;
  for(i=0; i<3 ; i++ ){
    featureVector[i] = *(longPtr++) ; // we should make this OS-independent
  }
  return 0 ;
}


//
// Given an image, computes the Example features (this is the API wrapper
// to the class specific stuff for this datatype)
//
int QbExampleFeatureClass::ComputeFeature(const QbGenericDataClass *data,
                                        const QbParameterClass *parms )
{
  QbGenericImageDataClass *imageData ;
  int rc = 0 ;
  int i ;

  // convert the generic data to the type we care about (image in this example)
  imageData = COERCE( data, QbGenericImageDataClass ) ;
  if( NULL == imageData ) return -1  ; // conversion failed

  // make sure we're in a good initial state
  for(i=0; i<3; i++) featureVector[i] = 0 ;  

  // we might have to do something different depending on the type of
  // data that was passed
  switch ( imageData->GetType() ) {
    // memory and picker usually mean we getting raw data so this is
    // where we do the data-specific feature calculation;
    // we might want to treat picker data differently, e.g., picker
    // image data is typically a very simple (non-dithered) image
    // and that might mean we do something different
    
  case QbGenericDataClass::memory:
    // put real code here for computing the features on a real image
    rc = 0 ; // means the feature computer was successful
    break ;

  case QbGenericDataClass::picker :
    // put real code here for computing the features on a picker image
    rc = 0 ; // means the feature computer was successful
    break ;

  case QbGenericDataClass::sample:
  case QbGenericDataClass::example:
  // in these cases, the data we want is in the database, so we just
  // go get it and convert it
    rc = FromByteString( imageData->GetData() ) ;
    break ;
  default:
    return -1 ; // oops, data wasn't a recognized type
  }

  // if this is a query and our feature computation is in a good
  // state, we may want to do some other stuff here (e.g.,
  // read the parameters and set the state of the distance
  // function)
  if ( imageData->GetIsQuery() && rc == 0 ) {
    // put the real post-processing code here
    rc = 0 ; // means any post processing for query worked OK
  }
  
  return rc ;
}

//
// Computes the distance from this feature to another feature
//
fastfloat QbExampleFeatureClass::Distance(const QbFeatureClass *feature,
                                        const QbParameterClass *parms,
                                        const float abortDistance )
{

  QbExampleFeatureClass *cf  ;
  // convert the generic feature class to this class 
  cf = COERCE( feature, QbExampleFeatureClass ) ;
  // if it didn't work, we may try SubPart
  if ( NULL == cf ) { // xxxSubPartFeatureClass has all needed interfaces
    cf = (QbExampleFeatureClass *)COERCE(feature, QbExampleSubPartFeatureClass);
  }
  // now check again
  if( NULL == cf ) { 
    QbError(QERR_BAD_COERCE2, QERR_NON_FATAL,
            "QbExampleFeatureClass::Distance", "QbFeatureClass",
            "QbExampleFeatureClass", "QbExampleSubPartFeatureClass");
    return -1.0 ;  // this is a bad distance, which gets checked 
  }

  // compute the real distance from the features in this class
  // and the features in the incoming class (cf) 

  return 0.0 ; // return real value of the distance 

}

float QbExampleFeatureClass::
ExecutionCostForDistanceFunction(const QbFeatureClass *feature,
                                 const QbParameterClass *parms,
                                 const float abortDistance )
{
  return -1.0 ;
}

float QbExampleFeatureClass::
ExecutionCostForFilterFunction(QbDatabaseClass &database,
                               QbKeyDatabaseClass *list,
                               const unsigned long queryResultsRequested,
                               const char *catalogName,
                               const QbParameterClass *parms )
{
  return -1.0 ;
}

int
QbExampleFeatureClass::FilterResult(QbDatabaseClass &database,
                                  QbKeyDatabaseClass *list,
                                  const unsigned long queryResultsRequested,
                                  const char *catalogName,
                                  const QbParameterClass *parms,
                                  QbKeyDatabaseClass &results )
{
  int returncode = -1 ; // default is we didn't do anything

#ifdef INDEX_CODE
  // usually we set numberOfListsRequested to about 20% of the dimension (d)
  // so we get about 20% of the database when the database is smaller
  // than d^2, but this example class has d=3 so that won't work.
  // We pick some reasonable number here (gives 20% of the database of size
  // 10,000) -- this should really be determined by experiment
  unsigned long numberOfListsRequested = 20 ;
  QbIndexClass indexer ;
  if ( parms->GetFilterFlag() ) {
    returncode = indexer.IndexFilter(database,
                                     *this,
                                     list,
                                     queryResultsRequested,
                                     catalogName,
                                     parms,
                                     numberOfListsRequested,
                                     results);
  }
#endif   
  return returncode ; 
}


int QbExampleFeatureClass::ToAsciiString( QbDatumClass &string )
{
  // this pretty-prints the data 
  unsigned char *buffer ;
  unsigned long size = 1024 ; // some number big enough to hold everything

  buffer = (unsigned char *)QbNew(size, "QbExampleFeatureClass::ToAsciiString");
  if (!buffer) return -1;
  sprintf((char *)buffer, "%9ld %9ld %9ld",
          featureVector[0], featureVector[1], featureVector[2] ) ; 
  // we made the storage here
  string.Set( size, buffer, QbDatumClass::newArrayStorage ) ; 
  return 0;
}
