/*
 *  Licensed Materials - Property of IBM
 *
 *  (C) COPYRIGHT IBM Corp. 1994-1998 All rights reserved
 *
 *  US Government Users Restricted Rights - Use, duplication or
 *  disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
 *
 * Filename : %M%
 * Date     : %G% %U%
 * Version  : %I%
 * Author   : Xiaoming Zhu (based on the work by Flick and Jim Hafner)
 * Purpose  : A high level Wrapper class for commandline programs such as
 *            QbMkDbs, QbQBE, QbMkThmb, QbDumpDb etc.
 *
 */

static char sccsid[] = "%W% %G%" ; 

static const char Copyright[] = {
  "Licensed Materials - Property of IBM\n"
  "(C) Copyright IBM Corp. 1994-1998  All rights Reserved\n"
  "US Government Users Restricted Rights - Use, duplication or\n"
  "disclosure restricted by GSA ADP Schedule Contract with IBM Corp." } ;

/* End of Common Block                                                  */

#include "QbicWrap.hpp"
#include <limits.h>
#include <time.h>

char *getOption(int *argc, char ***argv, char opt, int noOption = 0)
/* -------------------------------------------------------------------
  Author: Xiaoming Zhu
  Date:   8/23/98
  Purpose:
  Get command line option by opt flag.  If the flag is found, the option
  is returned.  Otherwise return NULL.  If noOption = 1 and opt flag is
  found, return a space (" ").
  Both option flag and/or option are removed from argc and argv.
---------------------------------------------------------------------*/
{
int ict, jct = 0;
char *flag=NULL, *cat = NULL, *tmp;

  for( ict=1; ict < *argc; ict++ ) {
    if( ('-' == (*argv)[ict][0]) && ((*argv)[ict][1] == opt)) {
       if(noOption) {
          if(ict > 1) {
            (*argv)[ict] = (*argv)[1];
            (*argv)[1] = (*argv)[0];
          }
          *argc -= 1;
          (*argv) += 1; 
          return " ";
       }
       jct = ict;
       flag = (*argv)[ict];
       if((ict+1)<*argc) cat = (*argv)[++ict];
       else return NULL;
       break;
    }
  }
  if(jct < 1) return NULL;
  else if (jct < (*argc) - 1) {
    for(ict=jct+2; ict<*argc; ict+=2,jct+=2) {
       (*argv)[jct] = (*argv)[ict];
       if(ict+1 < *argc) (*argv)[jct+1] = (*argv)[ict+1];
    }
  }
  *argc -= 2;
  return cat;
}

// default constructor to initialize the object to its default value.
QbicWrapClass::QbicWrapClass () 
{
   QbicWrapInit();
}

// overloaded constructor to initialize the object based on the input
// catalogName, databaseName, and connection mode.
QbicWrapClass::QbicWrapClass (char *catName, char *dirName, char *in_mode)
{
   QbicWrapInit();
   if( strcmp(catalogName, catName) ) strcpy(catalogName, catName);
#ifdef __MWERKS__
  {
    char tmp[256];
    Unix2MacDir(dirName, tmp);
    if(strcmp(databaseName, tmp)) strcpy(databaseName, tmp);
  }
#else
   if(strcmp(databaseName, dirName)) strcpy(databaseName, dirName);
#endif
   if( strcmp(mode, in_mode) ) strcpy(mode, in_mode);
}

// determining if the command line string contains a subport information
// which contains a part of " ( " and " ) ".
QbBoolean QbicWrapClass::isSubPart(char *cmdLine)
{
  if(strstr(cmdLine, " ( ") && strstr(cmdLine, " ) ")) return True;
  else return False;
}

// connect to database, default to read-only
int QbicWrapClass::ConnectToDatabase() {
   return ConnectToDatabase(databaseName, "r");
}

// connect to database named dirName with connection mode in_mode
int QbicWrapClass::ConnectToDatabase(char *dirName, char *in_mode) {
  int ict;
  if(NULL==dirName) {
    fprintf(stderr, "database name is not specified\n");
    return -1;
  }
#ifdef __MWERKS__
  {
    char tmp[256];
    Unix2MacDir(dirName, tmp);
    if(strcmp(databaseName, tmp)) strcpy(databaseName, tmp);
  }
#else
  if(strcmp(databaseName, dirName)) strcpy(databaseName, dirName);
#endif
  if(strcmp(mode, in_mode)) strcpy(mode, in_mode);
  if(connection) QbDestroyConnection(connection);
  connection = QbGetConnection(mode, databaseName);
  if(!connection) {
    printf("Error=%d  Severity=%d\n%s \n",
            GetErrorMSGNumber(), GetErrorMSGSeverity(), GetErrorMSG() ) ;
    return -1;
  }
// Create catalog class.  Catalog should be created from the connect object
// which has database name and connection mode.
  catalog = connection->CreateCatalogClass();
  if(!catalog) {
    QbDestroyConnection(connection);
    printf("Error=%d  Severity=%d\n%s \n",
            GetErrorMSGNumber(), GetErrorMSGSeverity(), GetErrorMSG() ) ;
    return -1;
  }
  if(catNeedOpen) {
    if(catalog->OpenCatalog(catalogName)) {
      catNeedOpen = 0;
      delete catalog;
      QbDestroyConnection(connection);
      return -1;
    }
  }
// If feature classes were created ahead of connection was made, featNeedAdd
// flag is on.
  if(featNeedAdd && mode[0]=='w') {
    for(ict=0; ict<numberOfFeatures; ict++) {
      catalog->AddFeatureClass( featNames[ict] ) ;
    }
  }
  return 0;
}

int QbicWrapClass::OpenCatalog(char *catName) {
/* -------------------------------------------------------------------------
   OpenCatalog: attempts to open the named catalog, if database connection
   is already there, 
   If no datatabase connection made yet, it simply copies the catalog name
   to the class and turn the catNeedOpen flag on.  The catalog will be added
   later when the database connection is made
-------------------------------------------------------------------------- */
  int ict;
  if(strcmp(catalogName, catName)) strcpy(catalogName, catName);
  if(!connection) {
    catNeedOpen = 1;
    return 1;
  }
  if(!catalog) {
    catalog = connection->CreateCatalogClass();
    if(!catalog) {
      QbDestroyConnection(connection);
      printf("Error=%d  Severity=%d\n%s \n",
            GetErrorMSGNumber(), GetErrorMSGSeverity(), GetErrorMSG() ) ;

      return -1;
    }
  }
  if(catalog->OpenCatalog(catName)) {
    delete catalog;
    QbDestroyConnection(connection);
    printf("Error=%d  Severity=%d\n%s \n",
            GetErrorMSGNumber(), GetErrorMSGSeverity(), GetErrorMSG() ) ;
    return -1;
  }

  if(featNeedAdd && mode[0] == 'w') {
    for(ict=0; ict<numberOfFeatures; ict++) {
      catalog->AddFeatureClass( featNames[ict] ) ;
    }
  }
  return 0;
}

int QbicWrapClass::CreateCatalog(char *catName) {
  int ict;
  if(strcmp(catalogName, catName)) strcpy(catalogName, catName);
  return catalog->CreateCatalog(catName);
}

int QbicWrapClass::QbicWrapClassConnect(char *catName, char *dirName,
                                        char *in_mode) {

  if(NULL==catName || NULL==dirName) {
    fprintf(stderr, "catalog name or database name unspecified catalogName=%s databaseName=%s\n", catName, dirName);
    return -1;
  }
#ifdef __MWERKS__
  {
    char tmp[256];
    Unix2MacDir(dirName, tmp);
    if(strcmp(databaseName, tmp)) strcpy(databaseName, tmp);
  }
#else
  if(strcmp(databaseName, dirName)) strcpy(databaseName, dirName);
#endif
  
  if(strcmp(catalogName, catName)) strcpy(catalogName, catName);
  if(strcmp(mode, in_mode)) strcpy(mode, in_mode);
  if(connection) QbDestroyConnection(connection);
  connection = QbGetConnection(mode, databaseName);
  if(!connection) {
    return -1;
  }
  if(catalog) delete catalog;
  catalog = connection->CreateCatalogClass();
  if(!catalog) {
    QbDestroyConnection(connection);
    return -1;
  }
  if(!strcmp(in_mode, "w")) CreateCatalog(catName);
  if(catalog->OpenCatalog(catName)) {
    delete catalog;
    QbDestroyConnection(connection);
    return -1;
  }
  return 0;
}

int QbicWrapClass::QbicWrapClassConnect(void) {
  if(NULL==catalogName || NULL==databaseName) {
    fprintf(stderr, "unknown catalogName=%s or databaseName=%s\n", 
                     catalogName, databaseName);
    return -1;
  } else return QbicWrapClassConnect(catalogName, databaseName, "r");
}

int QbicWrapClass::InsertFeatures(char *featureSpec)
{
  char *colon=NULL;
  char f[200], P[200];
  int rc=0;
  QbFeatureClass *thisFeature;
  QbParameterClass *thisParm;

  if( numberOfFeatures >= NUMBER_OF_FEATURES) {
    fprintf(stderr, "Exceed Max # (%d) of features\n", NUMBER_OF_FEATURES);
    return 1;
  }

  strcpy((char *)f, featureSpec);   // make a copy for safe keeping
  colon = strchr(f, ':');           // colon separates feature from parameter
  if ( colon ) {
    f[colon-f] = '\0'; // replace colon with null
    colon++; // step past colon, this marks the parameters
    strcpy((char *)P, colon);
  } else P[0] = '\0';

  thisFeature = QbCreateFeatureClass(f);
  if ( thisFeature ) { // try to construct and copy parameters
    thisParm = QbCreateParameterClass(thisFeature->ParameterClassname());
    if ( !thisParm ) { delete thisFeature; return -1; }
    *thisParm = *(thisFeature->GetDefaultParameters()); // copy ???
    if ( !thisParm->IsOK()) { //oops, cleanup
      delete thisParm;
      delete thisFeature;
      return -1;
    }
    if ( colon ) {
      rc = thisParm->FromCommandLineString(P);
      if (rc) {
        if (GetErrorMSGSeverity() == QERR_FATAL) {
          delete thisParm;
          delete thisFeature;
        }
        fprintf(stderr, "Skipping bad parameter in string %s for feature %s\n",
                P,f);
      }
    }
    strcpy(featNames[numberOfFeatures], thisFeature->Type());
    featureClass[numberOfFeatures] = thisFeature;
    parmsClass[numberOfFeatures++] = thisParm;
    if(catalog && mode[0]=='w') catalog->AddFeatureClass(thisFeature->Type());
    else featNeedAdd = 1;
  } else {
    fprintf(stderr, "feature %s was not created\n", f);
    return -1; // let caller decide what to do
  }
  return 0;
}

int QbicWrapClass::DeleteFeatures(char *featureName)
{
  int ict;
  for(ict=0; ict<numberOfFeatures; ict++) {
     if(!strcmp(featNames[ict], featureName)) {
        delete featureClass[ict];
        numberOfFeatures--;
        featureClass[ict] = featureClass[numberOfFeatures];
        parmsClass[ict] = parmsClass[numberOfFeatures];
        strcpy(featNames[ict], featNames[numberOfFeatures]);
        return 0;
     }
  }
  return 1;
}

int QbicWrapClass::GetParameters(QbStringClass *str)
{
  int ict;
  QbDatumClass datum;
  char *buffer;
  unsigned long size;
  for(ict=0; ict<numberOfFeatures; ict++) {
     parmsClass[ict]->ToCommandLineString(datum);
     datum.Get(size, buffer);
     *str += featureClass[ict]->Type();
     *str += ":";  
     *str += buffer;
     *str += "\n";
  }
  return 0;
}

QbParameterClass *QbicWrapClass::GetParamForFeature(char *featureName)
{
  int ict;
  for(ict=0; ict<numberOfFeatures; ict++) {
     if(!strcmp(featNames[ict], featureName)) {
        return parmsClass[ict];
     }
  }
  return NULL;
}

int QbicWrapClass::ProcessInputFileLine( char *inputFileLine, 
                                    QbBoolean remove)
{
  int rc=0;
  QbBoolean subparts = isSubPart(inputFileLine);

  if(subparts) {
     if(remove) rc = QbicWrapDeleteSubImage(inputFileLine);
     else       rc = QbicWrapInsertSubImage(inputFileLine);
  } else {
     if(remove) rc = QbicWrapDeleteImage(inputFileLine);
     else       rc = QbicWrapInsertImage(inputFileLine);
  }
  return rc;
}

char *QbicWrapClass::parseParentKey(char *imgNames)
{
  char *key = imgNames;

  while( isspace(*key) && '\0' != *key) key++;
  if(key[0] == '(' || key[0] == '\0') return NULL;
  else return key;
}

char *QbicWrapClass::parseNextChild(char *imgNames)
{
  char *key = imgNames;

  while( isspace(*key) && '\0' != *key) key++;
  if(*key == ')' || *key == '\0') return NULL;
  else if(*key == '(') return parseNextChild(key+1);
  else return key;
}

int QbicWrapClass::QbicWrapInsertImage(char *imgNames) {
  int rc = 0, size;
  char *p = imgNames;

  while( (p=parseParentKey(p)) ) {
    size = QbNextWordSize(p);
    p[size] = '\0';
#ifdef __MWERKS__
    {
       char tmp[256];
       Unix2MacKey(p, tmp);
       p = tmp;
    }
#endif
    if (image.ReadImageDataFromFile(p)) return -1;
    rc = catalog->AddRecord( p, image, (const QbFeatureClass **) featureClass,
                             (const QbParameterClass **) parmsClass, 
                             numberOfFeatures, update);
    if(rc) return rc;
    else p += size;
  }
  return rc;
}

int QbicWrapClass::QbicWrapDeleteImage(char *imgNames)
{
  int rc, size;
  char *p = imgNames;

  while( (p=parseParentKey(p)) ) {
    size = QbNextWordSize(p);
    p[size] = '\0';
#ifdef __MWERKS__
    {
       char tmp[256];
       Unix2MacKey(p, tmp);
       p = tmp;
    }
#endif
    rc = catalog->DeleteRecord( p );
    if(rc) return rc;
    else p += size+1;
  }
  return rc;
}

int QbicWrapClass::QbicWrapInsertSubImage(char *imgNames) {
// imgNames = "imgParent ( subimg1 subimg2 ... subimgN )"
// or       = "( subimg1 subimg2 ... subimgN )"

  int rc=0, size;
  char *p = imgNames, *q;
  char inputFileName[512];
 
  // Insert the Parent
  
  p = parseParentKey(p);
  if( p ) {
    size = QbNextWordSize(p);
    p[size] = '\0';
#ifdef __MWERKS__
    {
       char tmp[256];
       Unix2MacKey(p, tmp);
       p = tmp;
    }
#endif
    strcpy(inputFileName, p);
    rc = QbicWrapInsertImage(p);
    if(rc) return rc;
  }

  p[size] = ' ';
  p += size + 1;
  while( rc == 0 && (p = parseNextChild(p)) ) {
    size = QbNextWordSize(p);
    p[size] = '\0';
    q = p;
#ifdef __MWERKS__
    {
       char tmp[256];
       Unix2MacKey(q, tmp);
       q = tmp;
    }
#endif
    if (image.UpdateMaskInImageObject(q)) return -1;
    rc = catalog->AddRecord( q, image, (const QbFeatureClass **) featureClass,
                             (const QbParameterClass **) parmsClass, 
                             numberOfFeatures, update);
    if ( !rc ) catalog->DefineSubPart(inputFileName, q, image);
    p[size] = ' ';
    p += size+1;
  }
  return 0;
}

int QbicWrapClass::QbicWrapDeleteSubImage(char *imgNames) {
// imgNames = "imgParent ( subimg1 subimg2 ... subimgN )"
// or       = "( subimg1 subimg2 ... subimgN )"
  int rc=0, size;
  char *p = imgNames, *q;
 
  while( rc == 0 && (p = parseNextChild(p)) ) {
    size = QbNextWordSize(p);
    p[size] = '\0';
    q = p;
#ifdef __MWERKS__
    {
       char tmp[256];
       Unix2MacKey(q, tmp);
       q = tmp;
    }
#endif
    rc = catalog->DeleteSubPartRecord(q) ;
    p += size+1;
  }
  return 0;
}

int QbicWrapClass::GetFeatureDataString(char *string, QbStringClass *str)
{
  QbDatumClass datum;
  unsigned long size;
  char         *buffer;
  int ict, rc;

  QbicWrapQueryString(string);
  for(ict=0; ict<numberOfFeatures; ict++) {
     rc = queryImageOrKey[ict]->FromQueryString(featureClass[ict]->Type());
     rc = featureClass[ict]->ComputeFeature(queryImageOrKey[ict], 
                                            parmsClass[ict]);
     featureClass[ict]->ToAsciiString(datum);
     datum.Get(size, buffer);
     *str += featureClass[ict]->Type();
     *str += "\n";
     *str += buffer;
     *str += "\n";
  }
  return 0;
}

int QbicWrapClass::GetFeatureDataPicker(char *keyName, char *maskName,
                                       QbStringClass *str)
{
  QbDatumClass datum;
  unsigned long size;
  char         *buffer;
  int ict, rc;

  QbicWrapQueryPicker(keyName, maskName);
  for(ict=0; ict<numberOfFeatures; ict++) {
     rc = queryImageOrKey[ict]->FromQueryString(featureClass[ict]->Type());
     rc = featureClass[ict]->ComputeFeature(queryImageOrKey[ict], 
                                            parmsClass[ict]);
     featureClass[ict]->ToAsciiString(datum);
     datum.Get(size, buffer);
     *str += featureClass[ict]->Type();
     *str += "\n";
     *str += buffer;
     *str += "\n";
  }
  return 0;
}

int QbicWrapClass::GetFeatureDataImage(char *keyName, char *maskName,
                                       QbStringClass *str)
{
  QbDatumClass datum;
  unsigned long size;
  char         *buffer;
  int ict, rc;

  QbicWrapQueryImage(keyName, maskName);
  for(ict=0; ict<numberOfFeatures; ict++) {
     rc = queryImageOrKey[ict]->FromQueryString(featureClass[ict]->Type());
     rc = featureClass[ict]->ComputeFeature(queryImageOrKey[ict], 
                                            parmsClass[ict]);
     featureClass[ict]->ToAsciiString(datum);
     datum.Get(size, buffer);
     *str += featureClass[ict]->Type();
     *str += "\n";
     *str += buffer;
     *str += "\n";
  }
  return 0;
}

int QbicWrapClass::GetFeatureDataKey(char *keyName, QbStringClass *str)
{
  QbDatumClass datum;
  unsigned long size;
  char         *buffer;
  int ict, rc;

  QbicWrapQueryKey(keyName);
  for(ict=0; ict<numberOfFeatures; ict++) {
     rc = queryImageOrKey[ict]->FromQueryString(featureClass[ict]->Type());
     rc = featureClass[ict]->ComputeFeature(queryImageOrKey[ict], 
                                            parmsClass[ict]);
     featureClass[ict]->ToAsciiString(datum);
     datum.Get(size, buffer);
     *str += featureClass[ict]->Type();
     *str += "\n";
     *str += buffer;
     *str += "\n";
  }
  return 0;
}

char *QbicWrapClass::ListCatFeatures(void)
{
  int rc;
  unsigned long size;
  char          *buffer, *p;
  QbDatumClass featureList;
  if(!catalog) return NULL;
  rc = catalog->ListFeatureClasses(featureList);
  if(rc) return NULL;
  featureList.Get(size, buffer);
  p = (char *) QbNew(size, "QbicWrapClass::ListCatFeatures");
  if(!p) return NULL;
  strcpy(p, buffer);
  return p;
}

char *QbicWrapClass::ListAllFeatures(void)
{
  int ict;
  QbStringClass featList(128);
  unsigned long size;
  char          **buffer, *p, *q;

  buffer = (char **) QbListAllAvailableFeatures();
  if(!buffer) return NULL;
  for(ict=0; buffer[ict]; ict++) {
     featList += buffer[ict];
     featList += " ";
  }
  if(!featList.IsOK()) return NULL;
  q = (char *) featList;
  size = 1 + strlen(q);
  p = (char *) QbNew(size, "QbicWrapClass::ListAllFeatures");
  if(!p) return NULL;
  strcpy(p, q);
  return p;
}

int QbicWrapClass::QbicWrapCleanUp()
{
  int ict;

  if(catalog) delete catalog;
  if(connection) QbDestroyConnection(connection);
 
  for(ict=0; ict<numberOfFeatures; ict++) {
     delete featureClass[ict];
     delete parmsClass[ict];
  }
 
  for(ict=0; ict<numberOfFeatures; ict++) {
     if(queryImageOrKey[ict]) delete queryImageOrKey[ict];
  }
 
  for(ict=0; ict<numberOfSamplers; ict++) {
     delete sampleFeatureClass[ict];
     delete sampParmsClass[ict];
  }

  QbicWrapInit();
  return 0;
}

// Create a set of QbGenericImageDataClass objects, each for a given input
// feature class.
int QbicWrapClass::QbicWrapImageObj()
{
  int ict, jct;

  for(ict=0; ict<numberOfFeatures; ict++) {
     queryImageOrKey[ict] = (QbGenericImageDataClass *) 
                        QbCreateGenericDataClass("QbGenericImageDataClass");
     if(!queryImageOrKey[ict]) {
       for(jct=0; jct<ict; jct++) {
          delete queryImageOrKey[jct];
       }
       return -1;
     }
  }
  return 0;
}

// Create a set of QbGenericImageDataClass objects, initialize
// them with the image and mask in the file named imgName and maskName
int QbicWrapClass::QbicWrapQueryImage(char *imgName, char *maskName)
{
  int ict, rc;

  rc = QbicWrapImageObj();
  if(rc) return rc;

  for(ict=0; ict<numberOfFeatures; ict++) {
     rc = queryImageOrKey[ict]->ReadImageDataFromFile(imgName, maskName,
                                featureClass[ict]->IsSubPartFeature());
     if(rc) return rc;
  }
  return 0;
}

// Create a set of QbGenericImageDataClass objects, initialize
// them with the key named imgName.
int QbicWrapClass::QbicWrapQueryKey(char *imgName)
{
  int ict, rc;

  rc = QbicWrapImageObj();
  if(rc) return rc;

  for(ict=0; ict<numberOfFeatures; ict++) {
     rc = queryImageOrKey[ict]->MakeQueryByExampleObject(catalogName, imgName);
     if(rc) return rc;
  }
  return 0;
}

// Create a set of QbGenericImageDataClass objects, initialize
// them with the image description string named imgString.
int QbicWrapClass::QbicWrapQueryString(char *imgString)
{
  int ict, rc;

  rc = QbicWrapImageObj();
  if(rc) return rc;

  for(ict=0; ict<numberOfFeatures; ict++) {
     rc = queryImageOrKey[ict]->ReadPickerImageDescriptionString(imgString,
                                featureClass[ict]->IsSubPartFeature());
     if(rc) return rc;
  }
  return 0;
}

// Create a set of QbGenericImageDataClass objects, initialize
// them with the picker image and mask in the files named imgName and maskName
int QbicWrapClass::QbicWrapQueryPicker(char *imgName, char *maskName)
{
  int ict, rc;

  rc = QbicWrapImageObj();
  if(rc) return rc;

  for(ict=0; ict<numberOfFeatures; ict++) {
     rc = queryImageOrKey[ict]->ReadImageDataFromPickerFile(imgName, maskName,
                                featureClass[ict]->IsSubPartFeature());
     if(rc) return rc;
  }
  return 0;
}

// Query the database and output results in str.  It is assumed that the
// FeatureClass, image data classes, etc. have been created and initialized
// in previous method calls.
int QbicWrapClass::QbicWrapQueryDB(QbStringClass *str)
{
  int rc;
  QbGenericImageDataClass keywordData, *keywordPtr = &keywordData;
  QbFeatureClass *keywordClass;
  QbDbIteratorClass *iter = NULL;
  float distance;
  QbDatumClass key, value;
  char aline[800];
  unsigned long size;
  char *buffer; 
  const QbParameterClass *keywordParm;
  QbKeyDatabaseClass *firstPass = NULL, *finalPass = NULL;

  queryClass = (QbQueryServerClass *) QbCreateQueryClass("QbQueryServerClass");
  if(strlen(keyWord)) {
     keywordClass = QbCreateFeatureClass("QbTextFeatureClass");
     rc = keywordData.MakeQueryByTextObject(keyWord);
     keywordParm = keywordClass->GetDefaultParameters();
     if(rc) return rc;
     firstPass = queryClass->Evaluate(LONG_MAX,
                                      catalog,
                                      NULL,
                                      NULL,
                                      1,
                                      (QbGenericDataClass **) &keywordPtr,
                                      &keywordClass, 
                                      (QbParameterClass **) &keywordParm);
     delete keywordClass;
     if(!firstPass || queryClass->GetReturnCode()) {
       delete queryClass;
       return -1;
     }
  }
  if(numberOfFeatures) {
     finalPass = queryClass->Evaluate(nHits,
                                      catalog,
                                      (QbKeyDatabaseClass *) firstPass,
                                      NULL,
                                      numberOfFeatures,
                                      (QbGenericDataClass **) queryImageOrKey,
                                      featureClass,
                                      (QbParameterClass **) parmsClass);
  } else finalPass = firstPass;
  if(!finalPass) return -1;
  iter = finalPass->CreateIterator();
  if(!iter) return -1;
  sprintf(aline, "total of %ld keys found\n", finalPass->GetNumberOfKeys());
  *str += aline;

  while(iter && iter->More() ) {
    if(!iter->Next(key, value)) {
      key.Get( size, buffer );
      distance = finalPass->GetDistance(value);
      sprintf(aline, "%s %f\n", buffer, distance);
      *str += aline;
    }
  }

  delete iter;
  delete queryClass;
  queryClass = NULL;
  if(str->IsOK()) return 0;
  else return 1;
}

// Make a random query to the image database.  The output results will be
// stored in str.  If ranType is set true, the random results will be
// different each time this function is called.  If the ranType is set False,
// the return results are the same each time the function is called.

int QbicWrapClass::QbicWrapRandomQueryDB(QbStringClass *str, QbBoolean ranType)
{
  int ict, rc, num;
  int needFree = 0;
  unsigned long idx;
  float distance;
  char aline[800];
  unsigned long size;
  char *buffer, *p, *q; 
  time_t clock;

  QbFeatureClass *thisFeature;
  QbDatabaseClass *db = NULL;
  QbDbIteratorClass *iter = NULL;
  QbDatumClass key, value, distValue;
  QbStringClass tableName(64);
  QbKeyDatabaseClass *results = NULL;

  if(!connection || !catalog) {
    if(!connection) QbError(QERR_INFO, QERR_NON_FATAL, "No connection");
    if(!catalog)    QbError(QERR_INFO, QERR_NON_FATAL, "catalog not opened");
    return 1;
  }

  if(!numberOfFeatures) {
    rc = catalog->ListFeatureClasses(key);
    if(rc) {
      QbError(QERR_INFO, QERR_FATAL, "No Feature data in the catalog");
      return rc;
    }
    key.Get(size, buffer);
    p = buffer;
    while( (q=strchr(p, ' ')) != NULL) {
       *q = '\0';
       thisFeature = QbCreateFeatureClass(p);
       if(thisFeature) break;
    }
    if(!thisFeature) {
      QbError(QERR_INFO, QERR_FATAL, "Can't create any feature");
      return rc;
    } else needFree = 1;
  } else thisFeature = featureClass[0];
   
  
  num = catalog->NumberOfRecords (thisFeature->Type());
  if(!num) {
    if(needFree) delete thisFeature;
    QbError(QERR_INFO, QERR_FATAL, "No record in the feature table");
    return rc;
  }

  catalog->GetFeatureTableName(thisFeature->Type(), tableName);
  if(needFree) delete thisFeature;
  
  db = connection->CreateDatabaseClass();
  if(db) rc = db->Open(tableName);
  else {
    QbError(QERR_INFO, QERR_FATAL, "Cannot Create database class");
    return -1;
  }
  if(rc) {
    QbError(QERR_INFO, QERR_FATAL, "Cannot open feature table");
    return -1;
  }

  if(ranType) clock = time(&clock) | 1;
  else        clock = 1234;
  srand(clock);

  iter = db->CreateIterator();
  if(!iter) {
    QbError(QERR_INFO, QERR_FATAL, "Cannot create iterator");
    return -1;
  }
   
  results = connection->CreateKeyDatabaseClass();
  if(!results) {
    delete iter;
    QbError(QERR_INFO, QERR_FATAL, "Cannot create keydatabase");
    return -1;
  }

  while(iter->More()) {
    if(iter->Next(key, value)) {
      delete iter;
      delete results;
      return -1;
    }
    idx = (rand()<<16) + rand();

    idx %= num;
    distance = (float) idx / (float) num;
    if( results->SetDistance(distValue, distance) || 
        results->Insert(key, distValue) ) {
      delete results;
      delete iter;
      QbError(QERR_INFO, QERR_FATAL, "Cannot insert to keydatabase");
      return -1;
    }
  }

  db->Close();
  delete db;
  delete iter;
  iter = results->CreateIterator();
  if(!iter) {
     delete results;
     return -1;
  }
  rc = results->GetNumberOfKeys();
  if(rc > nHits) rc = nHits;
  sprintf(aline, "total of %d keys found\n", nHits);
  *str += aline;

  ict = 0;
  while( (ict++ < nHits) && iter->More() ) {
    if(!iter->Next(key, value)) {
      key.Get( size, buffer );
      distance = results->GetDistance(value);
      sprintf(aline, "%s %f\n", buffer, distance);
      *str += aline;
    }
  }

  delete results;
  delete iter;
  if(str->IsOK()) return 0;
  else return 1;
}

QbicWrapClass::~QbicWrapClass()
{
  QbicWrapCleanUp();
}

int QbicWrapClass::QbicWrapCDF(int *argc, char ***argv, char *mode)
{
  char *p;
  int rc=0;

  if( (p=getOption(argc, argv, 'd')) ) {
    rc = ConnectToDatabase(p, mode);
    if(rc < 0) {
       printf("Error=%d  Severity=%d\n%s \n",
            GetErrorMSGNumber(), GetErrorMSGSeverity(), GetErrorMSG() ) ;

      return -1;
    }
    else OK = True;
  } else return -1;

  if( (p=getOption(argc, argv, 'c')) ) {
    rc = OpenCatalog(p);
    if(rc < 0) {
      printf("Error=%d  Severity=%d\n%s \n",
            GetErrorMSGNumber(), GetErrorMSGSeverity(), GetErrorMSG() ) ;
      if(connection) QbDestroyConnection(connection);
      return -1;
    }
    else OK = True;
  } else return -1;

  p=getOption(argc, argv, 'f');
  while(p) {
    rc = InsertFeatures(p);
    if (rc) {
       printf("Error=%d  Severity=%d\n%s \n",
            GetErrorMSGNumber(), GetErrorMSGSeverity(), GetErrorMSG() ) ;
       if (GetErrorMSGSeverity() == QERR_FATAL) {
         QbicWrapCleanUp();
         return -1;
       }
    } else OK = True;
    p=getOption(argc, argv, 'f');
  }
  return 0;
}

// A wrapper method to index image based on various command line input.
// base for QbMkDbs

int QbicWrapClass::QbicWrapDBS(int argc, char **argv, QbStringClass *str)
{
  char *p, *q;
  int ict, jct, rc = 0, counter = 0;
  char inputFileLine[2000];
  FILE *fin;
  char formatline[80];

  if(argc < 2) {
     return QbMkDbsUsage(stdout);
  }

  if(QbicWrapCDF(&argc, &argv, "w")) return QbMkDbsUsage(stderr);
    
  if( (p=getOption(&argc, &argv, 'E', 1)) ) erase = True;
  if( (p=getOption(&argc, &argv, 'r', 1)) ) {
     remove = True; updateStr = "Deletion";
  }
  if( (p=getOption(&argc, &argv, 'i', 1)) ) {
    update = QbCatalogClass::incrementRecord;
    updateStr = "Incremental Update";
  }
  if( (p=getOption(&argc, &argv, 'u', 1)) ) {
    update = QbCatalogClass::updateRecord ;
    updateStr = "Update";
  }
  if( (p=getOption(&argc, &argv, 'U', 1)) ) {
    update = QbCatalogClass::forceUpdateRecord ;
    updateStr = "Update";
  }
  if( (p=getOption(&argc, &argv, 'l')) ) imageListName = p;
  while( (p=q=getOption(&argc, &argv, 'e')) ) {
    while(*q!='=' && *q!='\0') q++;
 
    if ( *q=='\0') { // oops! = is missing
      fprintf(stdout,"EnvVar=EnvValue pair missing = sign %s-- ignoring\n", p);
          break;
    }
    // set = to null and step past it with q
    p[q-p]='\0';
    // set the value -- bad return code means QbEnv doesn't know about key
    if ( Qbsetenv(p,q+1)!=0 ) fprintf(stdout,
                  "Environment variable %s unknown to Qbic -- ignoring\n", p);
    p[q-p]='=';
  }

  if(imageListName) { // Get image list from the file in imageListName
    fin = fopen(imageListName, "r");
    while(fin &&  (fgets(inputFileLine, 2000, fin) != NULL)) {
      rc = ProcessInputFileLine( inputFileLine, remove);
      if ( rc && GetErrorMSGSeverity() == QERR_FATAL) {
          fclose(fin); rc = -1;
      }
      counter++;
      strcpy(formatline, "<%d> %s of image/mask specs %s in database=%s ");
      if(rc) strcat(formatline, "catalog=%s unsuccessful due to %s\n");
      else   strcat(formatline, "catalog=%s successful\n");
      if(!str || SP) {
        if(rc) fprintf(stdout, formatline, counter, updateStr, inputFileLine,
                          databaseName, catalogName, GetErrorMSG() ) ;
        else   fprintf(stdout, formatline, counter, updateStr, inputFileLine,
                          databaseName, catalogName ) ;
      } else if(str) {
        char tmpline[500];
        if(rc) sprintf(tmpline, formatline, counter, updateStr, inputFileLine,
                       databaseName, catalogName, GetErrorMSG() ) ;
        else   sprintf(tmpline, formatline, counter, updateStr, inputFileLine,
                       databaseName, catalogName ) ;
        *str += tmpline;
      }
      fflush(stdout);
      fflush(stderr);
    }
    if(fin) fclose(fin);
  } else { // do the command line version
    for(ict=1; ict<argc; ict++) {
      inputFileLine[0] = '\0';
      if( strcmp(argv[ict], "+p") != 0) {
         strcat(inputFileLine, argv[ict]);
      } else {
         if(ict == argc - 1) break;
         if (strcmp(argv[++ict], "(") != 0) {
            strcat(inputFileLine, argv[ict++]);
            strcat(inputFileLine, " ");
         }
         for(jct=ict; jct<argc && argv[jct][0] != '-'; jct++) {
            strcat(inputFileLine, argv[jct]);
            strcat(inputFileLine, " ");
            if( argv[jct][0] == ')' ) { jct++; break; }
         }
         ict = jct - 1;
      }
      if( !strlen(inputFileLine) || strcmp(inputFileLine, "( )") == 0 )
          continue;
      counter++;
      rc = ProcessInputFileLine( inputFileLine, remove);
      strcpy(formatline, "<%d> %s of image/mask specs %s in database=%s ");
      if(rc) strcat(formatline, "catalog=%s unsuccessful due to %s\n");
      else   strcat(formatline, "catalog=%s successful\n");
      if(!str || SP) {
         if(rc) fprintf(stdout, formatline, counter, updateStr, inputFileLine,
                   databaseName, catalogName, GetErrorMSG() ) ;
         else   fprintf(stdout, formatline, counter, updateStr, inputFileLine,
                   databaseName, catalogName) ;
      } else if(str) {
         char tmpline[500];
         if(rc) sprintf(tmpline, formatline, counter, updateStr, inputFileLine,
                   databaseName, catalogName, GetErrorMSG() ) ;
         else   sprintf(tmpline, formatline, counter, updateStr, inputFileLine,
                   databaseName, catalogName ) ;
         *str += tmpline;
      }
    }
  }
  QbicWrapCleanUp();
  return rc;
}

// A wrapper method to make image query based on various command line input.
// base for QbQBE

int QbicWrapClass::QbicWrapQBE(int argc, char **argv, QbStringClass *str)
{
  int ict, rc = 0;
  char *maskFile = NULL, *buffer=NULL, *p, *q;
  QbStringClass strBuf(100);
  int idump = 0;

  if(argc < 2) {
     QbQBEUsage(stdout);
     return 0;
  }

  if( (getOption(&argc, &argv, 'h', 1)) ) return QbQBEUsage(stderr);

  if( (p=getOption(&argc, &argv, 'T')) ) strcpy(keyWord, p);
  if( (p=getOption(&argc, &argv, 'M')) ) maskFile = p;
  if( (p=getOption(&argc, &argv, 'n')) ) nHits = atoi(p);
  if( (p=getOption(&argc, &argv, 'P', 1)) ) idump = 1;
  if( (p=getOption(&argc, &argv, 'F')) ) {
    idump = 1;
    while(p) {
      rc = InsertFeatures(p);
      if(rc) {
        if (GetErrorMSGSeverity() == QERR_FATAL) { rc = -1; }
        break;
      } else OK = True;
      p=getOption(&argc, &argv, 'F');
    }
  }
  if( (p=getOption(&argc, &argv, 'L', 1)) ) {
    buffer = ListAllFeatures();
    if(NULL==buffer) {
      if(str) {
         *str = "Can't find any features for the api";
      } else fprintf(stderr, "Can't find any features for the api\n");
      rc = 1;
    }
    p = buffer;
    strBuf = "";
    while(p) {
      q = strchr(p, ' ');
      if(q) *q = '\0';
      strBuf += p;
      strBuf += "\n";
      if(q) p=q+1;
      else break;
    }
    if(SP) fprintf(stdout, "%s", (char *) strBuf);
    QbDelete(buffer);
    if(str) *str = (char *) strBuf;
    return 0;
  }
  if(QbicWrapCDF(&argc, &argv, "r")) return QbQBEUsage(stderr);
  if( (p=getOption(&argc, &argv, 'l', 1)) ) {
    buffer = ListCatFeatures();
    if(NULL==buffer) {
      strBuf = "Can't find any features for the api";
      fprintf(stderr, "%s\n", (char *) strBuf);
      if(str) *str = (char *) strBuf;
      return 1;
    }
    p = buffer;
    strBuf = "";
    while(p) {
      q = strchr(p, ' ');
      if(q) *q = '\0';
      strBuf += p;
      strBuf += "\n";
      if(q) p=q+1;
      else break;
    }
    if(SP) fprintf(stdout, "%s", (char *) strBuf);
    QbDelete(buffer);
    if(str) *str = (char *) strBuf;
    rc = 0;
  }

  if( (p=getOption(&argc, &argv, 'P', 1)) ) {
      rc = GetParameters(&strBuf);
      if(rc) {
        strBuf = "Error to get Parameters, have you added feature(s)?"; 
        fprintf(stderr, "%s\n", (char *) strBuf);
      }
      if(SP) fprintf(stdout, "%s\n", (char *) strBuf);
      if(str) *str = (char *) strBuf;
  } 
  if( (p=getOption(&argc, &argv, 'i')) ) {
      if(idump) rc = GetFeatureDataKey(p, &strBuf);
      else {
        rc = QbicWrapQueryKey(p);
        if(rc) {
          strBuf = "input key: "; 
          strBuf += p;
          strBuf += " does not exist in the database";
          fprintf(stderr, "%s\n", (char *) strBuf);
          QbicWrapCleanUp();
          if(str) *str = (char *) strBuf;
          return rc;
        }
        rc = QbicWrapQueryDB(&strBuf);
      }
      if(rc) {
        strBuf = "Error in database query with message: ";
        strBuf += GetErrorMSG();
        fprintf(stderr, "%s\n", (char *) strBuf);
        return rc;
      }
      if(SP) fprintf(stdout, "%s\n", (char *) strBuf);
      if(str) *str = (char *) strBuf;
  }
  if( (p=getOption(&argc, &argv, 'I')) ) {
      if(idump) rc = GetFeatureDataImage(p, maskFile, &strBuf);
      else {
        rc = QbicWrapQueryImage(p, maskFile);
        if(rc) {
          strBuf = "input image: "; 
          strBuf += p;
          strBuf += " does not exist!";
          fprintf(stderr, "%s\n", (char *) strBuf);
          QbicWrapCleanUp();
          if(str) *str = (char *) strBuf;
          return rc;
        }
        rc = QbicWrapQueryDB(&strBuf);
      }
      if(rc) {
        strBuf = "Error in database query with message: ";
        strBuf += GetErrorMSG();
        QbicWrapCleanUp();
        if(SP) fprintf(stderr, "%s\n", (char *) strBuf);
        if(str) *str = (char *) strBuf;
        return rc;
      }
      if(SP) fprintf(stdout, "%s\n", (char *) strBuf);
      if(str) *str = (char *) strBuf;
  }
  if( (p=getOption(&argc, &argv, 'S')) ) {
      if(idump) rc = GetFeatureDataString(p, str);
      else {
        rc = QbicWrapQueryString(p);
        if(rc) {
          strBuf = "input string: ";
          strBuf += p;
          strBuf += " in wrong format.";
          if(SP) fprintf(stderr, "%s\n", (char *) strBuf);
          QbicWrapCleanUp();
          if(str) *str = (char *) strBuf;
          return rc;
        }
        rc = QbicWrapQueryDB(str);
      }
      if(rc) {
        *str = "Error in database query with message: ";
        *str += GetErrorMSG();
        fprintf(stderr, "%s\n", (char *) *str);
      }
      if(SP) fprintf(stdout, "%s\n", (char *) *str);
  }
  if( (p=getOption(&argc, &argv, 'r', 1)) || 
      (q=getOption(&argc, &argv, 'R', 1)) ) {
      if(p) rc = QbicWrapRandomQueryDB(&strBuf);
      else  rc = QbicWrapRandomQueryDB(&strBuf, False);
      if(rc) {
         strBuf = "Error in database query with message: ";
         strBuf += GetErrorMSG();
         if(SP) fprintf(stderr, "%s\n", (char *) strBuf);
         if(str) *str = (char *) strBuf;
         return rc;
      }
      if(SP) fprintf(stdout, "%s\n", (char *) strBuf);
      if(str) *str = (char *) strBuf;
  }
  if(strlen(keyWord)) {
    rc = QbicWrapQueryDB(&strBuf);
    if(rc) {
      strBuf = "Error in keyword query with message: ";
      strBuf += GetErrorMSG();
      if(SP) fprintf(stderr, "%s\n", (char *) strBuf);
      if(str) *str = (char *) strBuf;
      return rc;
    }
    if(SP) fprintf(stdout, "%s\n", (char *) strBuf);
    if(str) *str = (char *) strBuf;
  }
  QbicWrapCleanUp();
  return rc;
}

// A wrapper method to make thumbnail images
// base for QbMkThmb

int QbicWrapClass::QbicWrapThm(int argc, char **argv, QbStringClass *str)
{
  int ict, rc = 0;
  char *imageFile = NULL, *imagePathList;
  char pathBuffer[1024];
  int idump = 0;
  char *p, *q, *outDir = NULL, *suffix = ".thumb0.gif";
  char inImg[256], outImg[256], keyName[256];
  QbStringClass strBuf(100);
  FILE *fin = NULL;

  if(argc < 2) {
     QbMkThmbUsage(stdout);
     return 0;
  }
  for( ict=1; ict < argc; ict++ ) {
    if( '-' == argv[ict][0] ) {
      switch ( argv[ict][1] ) {
      case 'c':
        thumb_24color = False;
        break ;
      case 'n':
        thumb_pad = False;
        break ;
      case 'p':
        thumb_pad = True;
        break ;
      case 't':
        thumb_24color = True;
        break ;
      case 'l':
        imageFile = argv[++ict];
        fin = fopen(imageFile, "r");
        if(!fin) {
          strBuf += "Image list file cannot be opened";
          if(SP) fprintf(stderr, "%s\n", (char *) strBuf);
          if(str) *str = (char *) strBuf;
          return -1;
        }
        break ;
      case 'h':
        thumb_y = atoi(argv[++ict]);
        break ;
      case 'w':
        thumb_x = atoi(argv[++ict]);
        break ;
      case 'o':
        outDir = argv[++ict];
        break ;
      case 's':
        suffix = argv[++ict];
        break ;
      case 'e':
        // general method for setting environment variables
        p=q=argv[++ict] ;
        // find the = sign but don't go off the deep end
        while(*q!='=' && *q!='\0') q++;
 
        if ( *q=='\0') { // oops! = is missing
          fprintf(stdout,"EnvVar=EnvValue pair missing = sign -- ignoring\n");
          break;
        }
        // set = to null and step past it with q
        p[q-p]='\0';
        q++;
        // set the value -- bad return code means QbEnv doesn't know about key
        if ( Qbsetenv(p,q)!=0 )
          fprintf(stdout,
                  "Environment variable %s unknown to Qbic -- ignoring\n", p);
        break ;
      default:
        break;
      }
    } else break;
  }

  if(fin) {
    imagePathList = Qbgetenv("QbicImagePath");
    if(imagePathList) {
#ifdef __MWERKS__
        Unix2MacDir(imagePathList, pathBuffer);
        imagePathList = pathBuffer;
#endif
    }
    while( fgets(keyName,255,fin) != NULL ) {
       keyName[QbNextWordSize(keyName)] = '\0'; // removes newline if exists
       rc = findfile(1, imagePathList, keyName, inImg, 199);
       if(outDir) {
          strcpy(outImg, outDir);
#ifdef __MWERKS__
          if(outImg[strlen(outImg)-1] != ':') strcat(outImg, ":");
#else
          strcat(outImg, "/");
#endif
       }
       strcat(outImg, keyName);
       strcat(outImg, suffix);
       rc = QbicWrapThumb(inImg, outImg);
       strBuf = "Create thumbnail ";
       strBuf += outImg;
       if(!rc) strBuf += " successful";
       else {
         strBuf += " failed due to ";
         strBuf += GetErrorMSG() ;
       }
       if(!str || SP) fprintf(stdout, "%s\n", (char *) strBuf);
       if(str) *str = (char *) strBuf;
    }
    fclose(fin);
  } else {
    strcpy(inImg, argv[ict++]);
    strcpy(outImg, argv[ict++]);
    rc = QbicWrapThumb(inImg, outImg);
    strBuf = "Create thumbnail ";
    strBuf += outImg;
    if(!rc) strBuf += " successful";
    else {
      strBuf += " failed due to ";
      strBuf += GetErrorMSG() ;
    }
    if(!str || SP) fprintf(stdout, "%s\n", (char *) strBuf);
    if(str) *str = (char *) strBuf;
  }
  QbicWrapCleanUp();
  return rc;
}

static QbBoolean IsBinary( const unsigned char *str, unsigned long size )
{
  int i ;
  // check all but last char
  for( i=0; i < (int) size-1; i++ )
    if( ! ( isprint( str[i] ) || isspace(str[i]) ) ) return True ;
  // None null terminated string - treat it as binary...
  if( str[size-1] != '\0' ) return True ;
  return False ;
}
static void PrintBinary( const unsigned char *str, unsigned long size, 
       FILE *stream )
{
  int i ;
  const char hexTable[] = "0123456789abcdef" ;
  for( i=0; i < (int) size; i++) {
    fprintf(stream, "%c%c", hexTable[ str[i] >> 4 ],
           hexTable[ str[i] & 0x0f ] ) ;
    if( i % 4 == 3 ) fprintf(stream, " ") ;
    if( i % 32 == 31 ) fprintf(stream, "\n") ;
  }
}

static int PrintFeatureData(QbDatabaseClass *db, QbFeatureClass *feature, 
       FILE *stream)
{
  QbDbIteratorClass *iter=NULL;
  QbDatumClass      key, value, asciiValue;
  unsigned long     keySize, valueSize;
  unsigned char     *keyChar, *valueChar;

  iter = db->CreateIterator();
  if(NULL==iter) {
      fprintf(stream, "Error=%d  Severity=%d\n%s \n",
            GetErrorMSGNumber(), GetErrorMSGSeverity(), GetErrorMSG() ) ;
      return -1;
  }
  while(iter->More()) {
    if (iter->Next( key, value)) { 
         fprintf(stream, "Error=%d  Severity=%d\n%s \n",
              GetErrorMSGNumber(), GetErrorMSGSeverity(), GetErrorMSG() ) ;
         break; }
    if ( feature ) {
      if ( feature->FromByteString( value ) ) {
        fprintf(stream, "Error=%d  Severity=%d\n%s \n",
            GetErrorMSGNumber(), GetErrorMSGSeverity(), GetErrorMSG() ) ;
        continue ;
      }
      if ( feature->ToAsciiString( asciiValue) ) {
        fprintf(stream, "Error=%d  Severity=%d\n%s \n",
            GetErrorMSGNumber(), GetErrorMSGSeverity(), GetErrorMSG() ) ;
        continue ;
      }
      asciiValue.Get( valueSize, valueChar ) ;
      key.Get( keySize, keyChar ) ;
      fprintf(stream, "%s %s\n", keyChar, valueChar ) ;
    } else {
      key.Get( keySize, keyChar ) ;
      value.Get( valueSize, valueChar ) ;

      if( IsBinary(keyChar, keySize) ) {
        fprintf(stream, "BKey[%d]:%c", keySize, keySize > 32 ? '\n': ' ' );
        PrintBinary( keyChar, keySize, stream ) ;
      } else printf("Key[%d]: %s", keySize, keyChar ) ;
      fprintf(stream, "\n") ;
      if( IsBinary(valueChar, valueSize) ) {
        fprintf(stream, "BData[%d]:%c", valueSize, valueSize > 32 ? '\n' : ' ') ;
        PrintBinary( valueChar, valueSize, stream ) ;
      } else  fprintf(stream, "Data[%d]: %s", valueSize, valueChar ) ;

      fprintf(stream, "\n\n") ;
    }
  }
  delete iter;
  return 0;
}

// A wrapper method to dump the content of a dbm database file
// base for QbDumpDb

int QbicWrapClass::QbicWrapDumpDb(int argc, char **argv, FILE *stream)
{
  int               ict, rc = 0;
  QbDatabaseClass   *db=NULL;
  QbStringClass     tableName(200);
  char              *catName, *dbName, *featName, *dbmFileName=NULL;
  char              tmp[256];

  if(argc < 2) {
     return QbDumpDbUsage(stdout);
  }

  catName = getOption(&argc, &argv, 'c');
   dbName = getOption(&argc, &argv, 'd');
 featName = getOption(&argc, &argv, 'f');
  while(featName) {
    InsertFeatures(featName);
    featName = getOption(&argc, &argv, 'f');
  }
  if(argc > 1) dbmFileName = argv[1];
  
  if(dbName) strcpy(tmp, dbName);
  else       strcpy(tmp, "./");

#ifdef __MWERKS__
  Unix2MacDir(tmp, databaseName);
#else
  strcpy(databaseName, tmp);
#endif
  connection = QbGetConnection("r", databaseName);
  if(!connection) {
     printf("Error=%d  Severity=%d\n%s \n",
            GetErrorMSGNumber(), GetErrorMSGSeverity(), GetErrorMSG() ) ;
     return -1;}
  db=connection->CreateDatabaseClass();
  if(!db)  {
      printf("Error=%d  Severity=%d\n%s \n",
            GetErrorMSGNumber(), GetErrorMSGSeverity(), GetErrorMSG() ) ;
      return -1;}
  if(catName) {
    catalog = connection->CreateCatalogClass();
    if(catalog) catalog->OpenCatalog(catName);
    else {
        printf("Error=%d  Severity=%d\n%s \n",
            GetErrorMSGNumber(), GetErrorMSGSeverity(), GetErrorMSG() ) ;
        return -1;}
    if(numberOfFeatures < 1) return QbDumpDbUsage(stdout);
    for(ict=0; ict<numberOfFeatures; ict++) {
      catalog->GetFeatureTableName(featureClass[ict]->Type(), tableName);
      rc = db->Open(tableName);
      if ( rc ) {
        printf("Unable to open %s. Skipping it. \n", (char *) tableName);
        continue ;
      }
      PrintFeatureData(db, featureClass[ict], stream);
      db->Close();
    }
  } else if(dbmFileName) {
    rc = db->Open(dbmFileName);
    if ( rc ) {
      printf("Unable to open %s. File may not exist.\n", (char *) dbmFileName);
      return -1;
    }
    if(numberOfFeatures < 1) featureClass[0] = NULL;
    PrintFeatureData(db, featureClass[0], stream);
    db->Close();
  }
  QbDestroyConnection(connection);
  QbicWrapCleanUp();
  return 0;
}
