/*
 *  
 *  Licensed Materials - Property of IBM
 *
 *  (C) COPYRIGHT IBM Corp. 1994-1998 All rights reserved
 *
 *  US Government Users Restricted Rights - Use, duplication or
 *  disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
 *
 * Filename : 76
 * Date     : 3/24/98 09:59:30
 * Version  : 1.4.3.4
 * Author   : Denis Lee
 * Purpose  : Class definition for the storage and import/export of parameters.
 *
 *
 */

/* End of Common Block                                               */

#ifndef __QbPrmCls_hpp
#define __QbPrmCls_hpp
#include <assert.h>
#include <dmbsystm.h>
////////////////////////////////////////////////////////////////
//
// Class to define parameters assocated with a feature/query
//
////////////////////////////////////////////////////////////////
class QbParameterClass : public QbBaseClass  {
private:
  QbParameterClass(const QbParameterClass&) ; // no copies
  enum { initsize = 16 } ;
 
protected:
  fastfloat classWeight; // class weight factor
  unsigned int count;    // # of features in the class
  unsigned int maxcount ; // max number of features we've got space for
  char  **featureCodes;   // a character value for the feature code
  char **featureCodeNames; // A character description of each class feature code
  fastfloat *weights;    // The floating point weight for each feature
  int  QueryOrder;    // determine how to run multifeature query
  int  CombineFunc;  // determine how to combine distances in multipass query.
                     // CombineFunc = 0: w1 * d1 + w2 * d2
                     // CombineFunc = 1: w2 * d2 (ignore distance in prior level
  Boolean doFilter ;    // controls whether FilterResult should be run
                        // within a given FeatureClass
private:
  //
  // member function to allocate more storage as needed.
  //
  int ReAllocateStorage(unsigned int i) {
    char **tmpFeatureCodes, **tmpFeatureCodeNames ;
    fastfloat *tmpWeights ;
    int oldmax = maxcount ;
    
    if ( IamOK == False ) return -1;
    if( i<maxcount) return 0 ;
    maxcount += initsize ;
    tmpFeatureCodes = new char *[maxcount];
    if (!tmpFeatureCodes) {
      QbError(QERR_MEM_WHERE, QERR_FATAL, maxcount,
              "QbParameterClass::ReAllocateStorage()");
      return -1;
    }
    tmpFeatureCodeNames = new char *[maxcount];
    if(!tmpFeatureCodeNames) {
      QbError(QERR_MEM_WHERE, QERR_FATAL, maxcount,
              "QbParameterClass::ReAllocateStorage()");
      delete [] tmpFeatureCodes;
      return -1 ;
    }
    tmpWeights = new fastfloat[maxcount];
    if(!tmpWeights) {
      QbError(QERR_MEM_WHERE, QERR_FATAL, maxcount,
              "QbParameterClass::ReAllocateStorage()");
      delete [] tmpFeatureCodes;
      delete [] tmpFeatureCodeNames;
      return -1;
    }
    memcpy( tmpFeatureCodes, featureCodes, sizeof( char *) * oldmax) ;
    memcpy( tmpFeatureCodeNames, featureCodeNames, sizeof( char *) * oldmax) ;
    memcpy( tmpWeights, weights, sizeof(fastfloat) * oldmax) ;
    
    memset( &(tmpFeatureCodes[oldmax]), 0, sizeof(char *) * initsize ) ;
    memset( &(tmpFeatureCodeNames[oldmax]), 0, sizeof(char *) * initsize ) ;
    memset( &(tmpWeights[oldmax]), 0, sizeof(fastfloat) * initsize ) ;
    delete [] featureCodes ;
    delete [] featureCodeNames ;
    delete [] weights ;
    featureCodes = tmpFeatureCodes ;
    featureCodeNames = tmpFeatureCodeNames ;
    weights = tmpWeights ;
    return 0 ;
  }
 
public:
  QbParameterClass() : classWeight(1.0), count(0), maxcount(initsize),
    featureCodes(NULL), featureCodeNames(NULL), weights(NULL),
    QueryOrder(0), CombineFunc(0), doFilter(False) {
      // allocate an initial segment of space
      unsigned int i;
      
      featureCodes = new char *[maxcount];
      if(!featureCodes) {
        QbError(QERR_MEM_WHERE, QERR_FATAL, maxcount,
                "QbParameterClass()");
        IamOK = False;
        return;
      }
      featureCodeNames = new char *[maxcount];
      if(!featureCodeNames) {
        QbError(QERR_MEM_WHERE, QERR_FATAL, maxcount,
                "QbParameterClass()");
        IamOK = False;
        return;
      }
      weights = new fastfloat[maxcount];
      if (!weights) {
        QbError(QERR_MEM_WHERE, QERR_FATAL, maxcount,
                "QbParameterClass()");
        IamOK = False;
        return;
      }
      
      for(i=0; i<maxcount; i++) {
        featureCodes[i] = NULL ;
        featureCodeNames[i] = NULL ;
        weights[i] = 0.0 ;
      }
  }
  
  virtual ~QbParameterClass() {
    unsigned int i ;
    for( i=0; i < count ; i++ ) {
      QbDelete(featureCodes[i])  ;
      QbDelete(featureCodeNames[i]) ;
    }
    delete [] featureCodes ;
    delete [] featureCodeNames ;
    delete [] weights ;
  }
  //
  // Copy assignment operator i.e. a = b, - copy all of the data
  //
  QbParameterClass& operator =(const QbParameterClass &rhs ) {
    unsigned int i ;
    SetClassWeight( rhs.GetClassWeight() ) ;
    SetFilterFlag( rhs.GetFilterFlag() ) ;
    SetQueryOrder( rhs.GetQueryOrder() ) ;
    SetCombineFunc( rhs.GetCombineFunc() ) ;
    for( i=0; i<rhs.count; i++ ) {
      // SetFeatureCodes sets IamOK=False on failure so we don't need to
      // check here to see if it worked (we're assuming anybody that
      // runs this copy will check later)
      SetFeatureCodes(rhs.featureCodeNames[i],rhs.featureCodes[i]);
      SetWeights(rhs.featureCodeNames[i], rhs.weights[i] ) ;
    }
    return *this ;
  }
  //
  // From QbBaseClass
  //
  virtual const QbInterfaceType IType() const
      { return ParameterInterfaceType ; }
 
  virtual const char *Type() const { return "QbParameterClass"; }

  int FromQueryString(const char *string) {
    QbDatumClass datum ;
    datum.Set( strlen((char *)string)+1, (unsigned char *) string ) ;  
    return FromQueryString( datum ) ;
  }
  
  // this could be simple: 'return ToCommandLineString'
  // if QbQueryS Create didn't look for '<' and '>' to mark the query string.
  // It used to be complicated, now it's a wrapper
  int ToQueryString( QbDatumClass &datum) const {
    QbDatumClass tmpDatum ;
    char *str ;
    char * buffer ;
    unsigned long size ;
    int rc ;
    // convert the features to a string in the datum
    rc = ToCommandLineString(tmpDatum) ;
    if ( rc ) return rc ;
    // get the datum into a buffer
    tmpDatum.Get(size, buffer) ;
    // get some space for the full string
    str = (char *) QbNew(size+10, "QbParameterClass::ToQueryString");
    if (!str) return -1;
    // wrap in < ... > pairs
    strcpy(str,"< ") ;
    strcat(str,buffer) ;
    strcat(str, " >") ;
    // set the datum
    datum.Set(strlen(str)+1, (unsigned char *)str, QbDatumClass::QbNewStorage) ;
    return 0 ;
  }
  
  // this used to be in the .cpp file. It's now a simple wrapper
  // As for ToQueryString, if it weren't for the '<' and '>' expected
  // by other functions, this would be simple
  int FromQueryString(const QbDatumClass &datum ){
    unsigned long size ;
    char *data, *string ;
    int rc ;
    datum.Get(size, data) ;
    string = (char *) QbNew(size, "QbParameterClass::FromQueryString");
    if (!string) return -1;
    sscanf(data, "< %s >", string) ;
    rc = FromCommandLineString(string) ;
    QbDelete (string) ;
    return rc ;
  }
 
  DMBEXPHP(int) FromCommandLineString(const QbDatumClass &datum ) {
    char * buffer ;
    unsigned long size ;
    datum.Get(size, buffer) ;
    return FromCommandLineString( buffer ) ;
  }
  // something for QbicMain to read command line specified parameters with
  DMBEXPHP(int) FromCommandLineString( const char *parmString ) ;
  // something to write for command line parsing (this probably
  // isn't needed, but the code here can be a model for things like
  // the java interface)
  DMBEXPHP(int) ToCommandLineString( QbDatumClass &datum ) const ;
  // something to get the parameters from a file
  DMBEXPHP(int) FromQueryStringInFile( const char *filename ) ;
 
  unsigned int GetCount() const  { return count ; }
 
  fastfloat GetClassWeight() const  { return classWeight; }
  void SetClassWeight( fastfloat x ) { classWeight = x ; }
  int GetQueryOrder() const  { return QueryOrder; }
  void SetQueryOrder( int x ) { QueryOrder = x ; }
  int GetCombineFunc() const  { return CombineFunc; }
  void SetCombineFunc( int x ) { CombineFunc = x ; }
  // get all feature codes
  char **GetFeatureCodes() const { return featureCodes ; }
  // get all feature code names
  char **GetFeatureCodeNames() const { return featureCodeNames ; }
  // get by name
  char *GetFeatureCodes(const char *name) const {
    for(unsigned int i=0;i<count; i++)
      if( strcmp(name,featureCodeNames[i]) == 0 )
        return featureCodes[i] ;
    // if we get through the loop nothing matched
    return (char *)NULL ;
  }
  // get by name
  int SetFeatureCodes( const char *name, const char *value ) {
    unsigned int i ;
    if ( IamOK == False ) return -1;
    for(i=0; i<count; i++)
      if( strcmp(name, featureCodeNames[i]) == 0 ) {
        QbDelete(featureCodes[i]) ;
        if (!(featureCodes[i] = QbStrDup(value))) { IamOK = False; return -1; }
        return 0;
      }
    // we didn't find it in the list so we add it, if we can room
    if (ReAllocateStorage(i)) { IamOK = False; return -1; }
    if (!(featureCodeNames[i] = QbStrDup(name)) ) { IamOK = False; return -1; }
    if (!(featureCodes[i] = QbStrDup(value))) {
      QbDelete(featureCodeNames[i]);
      IamOK = False;
      return -1;
    }
    count++ ; 
    return 0;
  }

  // get a pointer to the weights array
  fastfloat* GetWeights() const { return weights ; }

  // get by name
  fastfloat GetWeights(const char *name) const {
    for(unsigned int i=0;i<count; i++)
      if( strcmp(name,featureCodeNames[i]) == 0 )
        return weights[i] ;
    // if we get through the loop nothing matched
    return 0.0 ;
  }

  // set by name
  int SetWeights( const char *name, const fastfloat value ) {
    for(unsigned int i=0; i<count; i++)
      if( strcmp(name, featureCodeNames[i]) == 0 ) {
        weights[i] = value ;
        return 0;
      }
    return 1;
  }
 
  Boolean GetFilterFlag() const { return doFilter ; }
  void SetFilterFlag( const Boolean value ) { doFilter=value ; }
 
} ;
#endif
