/*
 *  
 *  Licensed Materials - Property of IBM
 *
 *  (C) COPYRIGHT IBM Corp. 1994-1998 All rights reserved
 *
 *  US Government Users Restricted Rights - Use, duplication or
 *  disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
 *
 * Filename : 37
 * Date     : 9/15/98 16:57:21
 * Version  : 1.2.1.1
 * Author   : Jim Hafner, Xiaoming Zhu, Myron Flickner
 * Purpose  : A master program to run all sorts of queries using the QBIC API.
 *            QbQBE runs two successive queries.  The first is a keyword
 *            query.  Its results are used as a restriction list for
 *            input to the second query.
 *
 */

/* End of Common Block                                               */

#ifndef __QbicMain_hpp
#define __QbicMain_hpp
#include "QbAbsCls.hpp"
#include "QbImgDat.hpp"
#include "QbCatCls.hpp"
#include "QbCreate.hpp"
static void Usage(char *pgm)
{
  const char usage1[] =
    "-Q querySpec\n\n"
    "\t-Q querySpec specifies a file to read for the query specification\n";
  const char usage2[] =
    "-h | -l | -L | -P | -r | -i|I|G|D query-image [-M mask]\n"
    "\t [-c catalog] [-d database] [-T keyword]\n"
    "\t [-f|F feature1[:parms1] [-f|F feature2[:parms2] ...]]\n"
    "\t [-n nresults] [-s] [-t] \n"
    "\t [-e EnvVar=EnvValue ]\n\n"
    "\t-h           help -- dump this usage message\n"
    "\t-l           the list of feature classes in the catalog (-c or default)\n"
    "\t-L           the list of feature classes known to Qbic\n"
    "\t-P           dump the default parameters for each of the -F features\n"
    "\t-r           specifies a \"random\" query -- returns a list of random\n"
    "\t             database keys in the feature (-f), catalog (-c or default)\n"
    "\t             and database (-d or default)\n"
    "\t-i dbkey     specify the database key (key must exist in db)\n"
    "\t             for the query\n"
    "\t-I image     specify the image file for the query\n"
    "\t-G image     specify the GUI-generated image file for the query\n"
    "\t-D image     specify the GUI-generated query image description file\n"
    "\t-S description  specify the GUI-generated query image description string\n"
    "\t-M mask      specify the image mask file for the query\n"
    "\t             (with -I or -G -- mask and image size must match)\n"
    "\t-c catalog   specifies the catalog name (defaults to \"test\")\n"
    "\t-d database  specifies the database name (defaults to \"QbicData\")\n"
    "\t-T keyword   specifies the single restriction-keyword to filter with\n"
    "\t             (first query)\n"
    "\t-f feature:parms\n"
    "\t             specifies the feature class(es) to use in the main (second)\n"
    "\t             query together with (optional) weights are parameters for\n"
    "\t             that feature -- see documentation for details on parameters\n"
    "\t-F feature   specifies the feature data to dump for -i,-I,-G,-D,-P options\n"
    "\t-n #         specifies the maximum number of results to return\n"
    "\t             (from second query) \n"
    "\t-e EnvVar=EnvValue\n"
    "\t             to set general Qbic environment variables\n"
#ifndef __MWERKS__
    "\t-s           run query (-i,-I,-G,-D, with -f only) on server\n"
#endif
    "\t-t           time the main query in milliseconds\n";
  printf("\n%s %s", pgm, usage1 );
  printf("\nOR\n");
  printf("\n%s %s", pgm, usage2 );
  //exit(200);
}
class QbicMainClass {
  enum QueryType { nullQuery, Random, catFeatureList, AllFeatureList, DBKey,
                   Image, GImage, Description, TextOnly, ParameterDump } ;
  enum runQueryRCType { OK, noFeaturesInQuery, noQuery, unsupported,
                        badImageData, noHits, noFeaturesInCat, noRecords,
                        noCatalog, noFeatureTable, imageQueryFailure,
                        textQueryFailure, noTextInQuery, fatal,
                        noConnection, noQueryInstance, badQuery, badArgs} ;
 
private:
#ifndef __MWERKS__
  char                    **argList ;
#else
  /* argvtype defined in QbMac.hpp, loaded by QbAbsCls.hpp */
  argvtype                argList;
#endif
  Boolean                 server, filter, TimeTest, IamOK ;
  QbStringClass           catalogName;
  const char              *keywordFeatureName;
  QbStringClass           queryImageName, queryMaskName, queryImageString;
  QbStringClass           aFeatureName, databaseName ;
  QbGenericDataClass      *keywordData;
  int                     argcount, nFeatures, nHits, numEnvs ;
  runQueryRCType          runQueryRC ;
  QueryType               thisQuery ;
  const char              *featureNames[NUMBER_OF_FEATURES] ;
  QbGenericImageDataClass *queryByExampleImage[NUMBER_OF_FEATURES];
  QbParameterClass        *parms[NUMBER_OF_FEATURES];
  const QbParameterClass  *defaultparms[NUMBER_OF_FEATURES];
  Boolean                 isSubPart[NUMBER_OF_FEATURES];
  int                     imageQueryRC ;
  char                    *EnvValues[25];
  Boolean                 doQueryFeature, doDataFeature;
  // called by the initializer to set the query specifications
  int setQuerySpecs() ;
  // the master arg parser (called by setQuerySpecs or parseQuerySpecFile)
  int parseArgList(char *argspec) ;
  // read the query specs from a file and convert to a form for parseArgList
  int parseQuerySpecFile( char *filename) ;
  // a method to dump default parameters
  void dumpParameters() ;
  // a method to get random keys from the database
  void getRandomKeys() ;
  // a method to get the features in a catalog
  void getCatalogFeatureList() ;
  // a method to add a feature class and parameters to featureNames and parms
  int AddQueryFeature(const char *featureName) ;
  // a method to get the data for the query items into the queryByExample array
  int packQueryByExampleImageData() ;
  // a method to run a typical image query
  void runImageQuery() ;
  // a method to dump the feature information for query object
  void dumpFeatureData() ;
 
public:
  QbicMainClass():
    // the Booleans
    server(False), filter(False), TimeTest(False), IamOK(True),
    // the non-null strings
    catalogName("test"),
    keywordFeatureName("QbTextFeatureClass"),
    // the null strings
    queryImageName((unsigned long)0), queryImageString((unsigned long)0),
    databaseName((unsigned long)0),
    queryMaskName((unsigned long)0), aFeatureName(64),
    // the datum
    keywordData(NULL),
    // the integers
    nFeatures(0), nHits(20), numEnvs(0),
    // the enums
    runQueryRC(OK), thisQuery(nullQuery),
    doQueryFeature(False), doDataFeature(False)
    {
      if (!catalogName.IsOK() || !queryImageName.IsOK() ||
          !queryMaskName.IsOK() || !aFeatureName.IsOK() || !databaseName.IsOK()) {
        IamOK = False;
      }
      for( int i=0;i<NUMBER_OF_FEATURES;i++) {
        featureNames[i] = NULL ;
        queryByExampleImage[i] = NULL ;
        parms[i] = NULL ;
        defaultparms[i] = NULL ;
        isSubPart[i] = False ;
      }
    } ;
 
  // this should cleanup after any query type
  virtual ~QbicMainClass()  ;
  // this is the guy that initializes the state from a command line
#ifndef __MWERKS__
  virtual void initialize(int argc, char **argv) ;
#else
  virtual void initialize(int argc, argvtype *argv);
#endif
  // this should run the query
  virtual void runQuery() ;
  // something to handle the return value from runQuery()
  virtual void runQueryRCHandler() ;
  // a method to return the runQueryRC as an int
  int getRunQueryRC() { return (int)runQueryRC ; } ;
  // a method for checking state of constructor
  Boolean IsOK() { return IamOK; };
} ;
#endif
