/*
 *  IBM Confidential
 *  
 *  Licensed Materials - Property of IBM
 *
 *  (C) COPYRIGHT IBM Corp. 1994, 1995, 1996, 1997  All rights reserved
 *
 *  US Government Users Restricted Rights - Use, duplication or
 *  disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
 *
 * Filename : genimage.h
 * Author   : Xiaoming Zhu, Jim Hafner
 * Date     : 2/10/98 17:24:19
 * Version  : 1.7
 * Purpose  : API's for General Image Library (wrappers to JPEG LIB and LIBTIFF)
 * */

/* static char sccsid[] = "@(#)genimage.h	1.7 2/10/98" ; */

#ifndef __GENIMAGE
#define __GENIMAGE
#ifdef __cplusplus
extern "C" {
#endif
#include <stdio.h>

typedef enum {
        GEN_FMT_AUTO = 0,       /* sort of like NULL, determine from filename*/
        GEN_FMT_JPEG,           /* JPEG format */
        GEN_FMT_BMP,            /* BMP format (Windows flavor) */
        GEN_FMT_GIF,            /* GIF format */
        GEN_FMT_OS2,            /* BMP format (OS/2 flavor) */
        GEN_FMT_PPM,            /* PPM (PBMPLUS format) */
        GEN_FMT_PGM,            /* PGM (PBMPLUS format) */
        GEN_FMT_PBM,            /* PBM (PBMPLUS format) -- not supported */
        GEN_FMT_RLE,            /* RLE format */
        GEN_FMT_TARGA,          /* Targa format */
        GEN_FMT_TIFF,           /* TIFF format */
        GEN_FMT_G4TIFF,         /* G4 TIFF format */
        GEN_FMT_UNKNOWN         /*  Unknown type, we choke on this */
} GENIMAGE_FORMATS;

typedef struct gen_img_struct * gen_img_ptr;

/* ------------------------------------------------------------------------
                                      grayscale or colormapped    24 bits
   in_components, out_components   =             1                  3   

             
   in_color_space, out_color_space = 0 (grayscale)  1 (colormapped or 24bits)

   in_img24, out_img24   [row][col*3];
   in_bmp, out_bmp       [row][col];
   in_lut, out_lut       [3][256];

   in_img_pre            in_img24 present 
   in_lut_pre            colormap present
   in_bmp_pre            in_bmp present (palettized!)

   in_fname              filename of image
   out_rotation          should be part of flags to determine transformations
   fin, fout             FILE pointers
   in_format             image type (jpeg, tif, etc)
   in_compress_ptr       place to put cinfo, dinfo when we de-globalize
   in_iomgr_ptr          place to put specific io manager (e.g., TIFF pointer)
------------------------------------------------------------------------ */
struct gen_img_struct {
   unsigned long in_width, in_height;
   unsigned long out_width, out_height;
   unsigned int  in_components, out_components;
   int           in_color_space, out_color_space;
   unsigned char **in_img24, **in_bmp, **in_lut;
   unsigned char **out_img24, **out_bmp, **out_lut;
   int           in_img_pre, in_bmp_pre, in_lut_pre;
   int           out_img_pre, out_bmp_pre, out_lut_pre;
   char          *in_fname, *out_fname;
   int           out_rotation;
   FILE          *fin, *fout;
   GENIMAGE_FORMATS in_format, out_format;
   void          *in_compress_ptr, *out_compress_ptr;
   void          *in_iomgr_ptr, *out_iomgr_ptr ;
   void          *in_err_ptr, *out_err_ptr ;
   char          in_type[10], out_type[10];  /* can deprecate this */
   int           in_is_jpeg, out_is_jpeg; /* can deprecate this */
};

/* ------------------------------------------------------------------------
 GEN_IMG_ERROR: error handling, 
         return status: non-zero=return bad state, 0=OK return
         msg: error message
 gen_img_error_mgr:  method to override the jpeg error_mgr structure
                     so we can override the error_exit to get more
                     graceful returns (not fatal exits)

 gen_img_error_exit: this does longjmp to return control for cleanup
                     rather than exit
                     
 write_jpeg_file: method to write a file with given name, using
                  requested quality compression, from rgb array
                  of bits (img24) of given size

 two_pass_palettize: converts img24 of given size to img8 and lut

 gen_img_packmask: converts a bilevel mask array to G4TIFF compressed
                   image array

 gen_img_unpackmask: converts G4TIFF compressed image array to bilevel
                   (0,255) mask

 gen_img_read_sizes: read the given file for enough header information
                     to set the WxH, closes if requested

 gen_img_get_nonpalettized: read images into nonpalettized array.

 gen_img_put_nonpalettized: write image nonpalettized.

 gen_img_get_palettized: starts with preinitialized ginfo (with open
                         already done, reads data, palettizes if necessary
                         and puts the data in imagedat(img8) and lutdat(lut)

 gen_img_put_palettized: takes imagedat(img8) and lutdat(lut) and writes
                         an image with the given filename (type is determined
                         from the filename).
                     
 return:
       0: success
   other: errors in processing.

 *-------------------------------------------------------------------------- */

extern int write_jpeg_file(char * filename, int quality, unsigned char *rgb,
                           unsigned long image_width, unsigned long image_height);

extern int two_pass_palettize(unsigned char **img24, unsigned long w,
      unsigned long h, unsigned char **img8, unsigned char **lut);

extern long gen_img_packmask(unsigned char *mask, char *packmask,
                            unsigned long xsize, unsigned long ysize) ;

extern int gen_img_unpackmask(unsigned char *mask, char *packmask,
                       int maskpacksize, unsigned long xsize, 
                       unsigned long ysize) ;

extern int gen_img_read_sizes(char *filename,
                              long *xSize, long *ySize,
                              gen_img_ptr ginfo, int close_when_done) ;

extern int gen_img_read_close(gen_img_ptr ginfo) ;

extern int gen_img_get_nonpalettized(gen_img_ptr ginfo,
                                  char *imagedat);

extern int gen_img_put_nonpalettized(char *filename,
                                    long xsize,
                                    long ysize,
                                    char *imagedat,
                                    int ImageType);

extern int gen_img_get_palettized(gen_img_ptr ginfo,
                                  char *imagedat,
                                  char *lutdat);

extern int gen_img_put_palettized(char *filename,
                                  long xsize,
                                  long ysize,
                                  char *imagedat,
                                  char *lutdat);

extern int gen_img_rescale_image(char *in_name, char *out_name,
                                 unsigned long w, unsigned long h, int pad);


#ifdef __cplusplus
}
#endif

#endif
